/*
 * Copyright (C) 2001 and 2002 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef _CTMPL_ITERATOR_H_
#define _CTMPL_ITERATOR_H_

	/* iterator macros */
#define iterator(x)		x##_iterator
#define iterator_clear(x)	memset(&x, 0, sizeof(x)) 
#define iterator_type(x)	((x).iter_type)
#define iterator_aggregate(x)	((x).iter_aggr)
#define iterator_item(x)	((x).iter_item)

/* subtype is used for implementation specific purpose */
#define iterator_subtype(x)	((x).iter_subtype)

/* 
 *  to get data contained by item 
 *	if iterator is end-iterator, just return (casted)-1.
 *	caller must check it out before call this method.
 *
 *	if(!iterator_is_end(itr)){
 *		data = iterator_data(itr);
 *	}
 */
#define iterator_data(x)	(((x).iter_aggr)->ctl->retrieve(x))  

/*  
 *  Data Structure imdependent interface for iterator.
 *	following interfaces returns iterator whose type is CTMPL_ITER_FORWARD.
 *	iterator(list) itr = iterator_first(list);
 *	iterator(list) itr = iterator_last(list);
 *	iterator(list) itr = iterator_at(list, 3);
 *	iterator(list) itr = iterator_at(list, -1);
 *
 *	if position indicated by iterator_at() is out of range, 
 *	end iterator returns. 
 *
 *	iterator_first(list) is same as iterator_at(list, 0);
 *	iterator_last(list) is same as iterator_at(list, -1);
 */
typedef int iterator_pos_t;
typedef unsigned int iterator_cnt_t;
#define iterator_first(x)	((x)->ctl->first)(x)
#define iterator_last(x)	((x)->ctl->last)(x)
#define iterator_end(x)		((x)->ctl->enditr)
#define iterator_at(x, y)	((x)->ctl->at)((x), (y))
#define iterator_count(x)	((x)->ctl->count)((x))

#define iterator_is_empty(x)	(iterator_count(x) == 0)
#define iterator_is_equal(x, y)	(iterator_aggregate((x))->ctl->is_equal)(x, y)

/* 
 *	iterator_is_end(x) should be 
 *	(iterator_type(x) == CTMPL_ITER_END) 
 *	but is_equal reuqires aggregators pointer. so 
 *	currently don't use iterator_end(iterator_aggregate(x)).
 */
#define iterator_is_end(x)   \
	(iterator_item(x) == \
		iterator_item(iterator_end(iterator_aggregate(x))))

/*  
 *   set iterator type.
 *	iterator_settype(list, CTMPL_ITER_BACKWARD);
 */
#define CTMPL_ITER_END		-1	
#define CTMPL_ITER_FORWARD	 0
#define CTMPL_ITER_BACKWARD	 1
#define iterator_settype(x, y)	(iterator_type(x) = y)
#define iterator_set_forward(x)	(iterator_settype(x, CTMPL_ITER_FORWARD))
#define iterator_set_backward(x) (iterator_settype(x, CTMPL_ITER_BACKWARD))

/*  
 *  return iterator of next/prev position.
 */
#define iterator_next(x)	\
	(iterator_type(x) == CTMPL_ITER_FORWARD ? \
		(((x).iter_aggr)->ctl->next(x)) : \
		(((x).iter_aggr)->ctl->prev(x)))

#define iterator_prev(x)	\
	(iterator_type(x) == CTMPL_ITER_FORWARD ? \
		(((x).iter_aggr)->ctl->prev(x)) : \
		(((x).iter_aggr)->ctl->next(x)))

#define iterator_go(x, y) \
	do {						   \
		int i = ((y) < 0? -(y): (y)), isminus = ((y) < 0); \
		while(i > 0){				   \
			x = (isminus? iterator_prev(x): iterator_next(x)); \
			--i; \
		} \
	} while(0)
		

/*  
 *   erase current item from the aggregation. 
 *  free memory space of the current item and return iterator of 
 *  next position.
 */
#define iterator_erase(x)	\
	(iterator_type(x) == CTMPL_ITER_FORWARD ? \
		(((x).iter_aggr)->ctl->erase_gonext(x)) : \
		(((x).iter_aggr)->ctl->erase_goprev(x)))

/*
 *   access for each iterator data 
 */
#define iterator_foreach(x, func, arg)	\
		((x)->ctl->foreach(x, func, arg))

/**************************************************************/
/***            Aggregator specific interfaces              ***/
/**************************************************************/
/* 
 *   find iterator by key. 
 */
#define iterator_find(x,y)   ((x)->ctl->find)((x), (y))

/* 
 *   find iterator by key with specific compare function. 
 */
#define iterator_find_by(x,y,z)   \
	((x)->ctl->find_between_by(iterator_first(x), \
				   iterator_end(x), (y), (z)))

/* 
 *   find iterator between indicated range.
 */
#define iterator_find_between(a,b,y)   \
	((iterator_aggregate(a))->ctl->find_between((a), (b), (y), NULL))

/* 
 *   find iterator between indicated range by specific cmp func.
 */
#define iterator_find_between_by(a,b,y,z)   \
	((iterator_aggregate(a))->ctl->find_between((a), (b), (y), (f)))

/* 
 *   insert data.
 */
#define iterator_insert(x, y) ((x)->ctl->insert)((x), (y))

/* 
 *   insert pair data(first/second)
 */
#define iterator_insert_pair(x, y, z) ((x)->ctl->insert_pair)((x), (y), (z))

/* 
 *   prepend/append data.
 */
#define iterator_prepend(x, y) \
	((x)->ctl->insert_before)(iterator_first(x), (y))

#define iterator_append(x, y) \
	((x)->ctl->insert_after)(iterator_last(x), (y))

/* 
 *   insert data before/after indicated iterator.
 */
#define iterator_insert_before(x, y) \
	(iterator_type(x) == CTMPL_ITER_FORWARD ? \
		(iterator_aggregate(x)->ctl->insert_before)((x), (y)):\
		(iterator_aggregate(x)->ctl->insert_after)((x), (y)))

#define iterator_insert_after(x, y) \
	(iterator_type(x) == CTMPL_ITER_FORWARD ? \
		(iterator_aggregate(x)->ctl->insert_after)((x), (y)):\
		(iterator_aggregate(x)->ctl->insert_before)((x), (y)))

/* 
 *   swap data indicated iterators.
 */
#define iterator_swap(x, y) \
		(iterator_aggregate(x)->ctl->swap)((x), (y))

/* 
 *   assign data into iterator's data.
 */
#define iterator_assign_at(a, p, v) \
		((a)->ctl->assign)(iterator_at(a, p), (v))
#define iterator_assign(i, v) \
		(iterator_aggregate(i)->ctl->assign)((i), (v))


/* c template internal Macro */
#define CTMPL_ITR_IS_EQUAL(x, y) \
	(iterator_aggregate(x) == iterator_aggregate(y) && \
	 iterator_item(x) == iterator_item(y)) 

#endif /* _CTMPL_ITERATOR_ */
