/*
 *			ndtpbook.c
 * $Id: ndtpbook.c,v 1.7 2004/01/12 07:07:35 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/stat.h>

#include "memory.h"
#include "tcp.h"
#include "bitmap.h"
#include "property.h"
#include "dictbook.h"
#include "booklist.h"
#include "ndtpbook.h"
#include "ndtpserver.h"
#include "subr.h"

#include "ndtpcache.c"

static int 
ndtpcache_cmp(const char *p1, const char *p2)
{ 
	return strcmp(p1, p2); 
}

static unsigned int 
ndtpcache_hash(const char *p1)
{ 
	int i;
	unsigned int ret = 0;

	for(i = 0; i < 4; ++i){
		if(p1[i] == '\0')
			break;
		ret += p1[i];
	}
	return ret;
}

/* cache entry */
struct cacheent {
	const char *nc_key;	/* word/ptr/fontptr */
	strlist	    nc_value;	/* value */
	struct bitmap *nc_bmp;	/* bitmap */
};

struct cacheent *
cacheent_new(const char *key, strlist val)
{
	struct cacheent *ret = MALLOC("cacheent", sizeof(struct cacheent));

	ret->nc_key = (const char *)xp_strdup(key);
	ret->nc_value = val;
	ret->nc_bmp = NULL;
	return ret;
}

void
cacheent_delete(const char *p, struct cacheent *ent)
{
	xp_strfree(ent->nc_key);
	strlist_delete(ent->nc_value);
	if(ent->nc_bmp)
		bitmap_free(ent->nc_bmp);
	FREE("cacheent", ent);
}

static void
remove_multislash(char *path)
{
	register char *p = path;

	while(*p != '\0'){
		if(*p == '/' && *(p + 1) == '/')
			strcpy(p, p+1);
		else
			++p;
	}

	if(path[strlen(path)-1] != '/')
		strcat(path, "/");
}

static int
cachedir_create(char *path)
{
	extern int errno;
	char *p = path;

	remove_multislash(path);
	++p;
	
	while((p = index(p, '/'))){
		*p = '\0';

		if(access(path, R_OK | X_OK) < 0){ 
			if(errno != ENOENT) 
				return -1;
			else if(mkdir(path, 0777) < 0)
				return -1;
		}
		*p = '/';
		++p;
	}
	return 0;
}


/* dump a cache entry on memory into disk */
static int 
cache_write(struct ndtpbook *nbp, struct cacheent *cp)
{
	FILE *fp;
	char path[MAXPATHLEN];
	strlist_iterator itr;

	snprintf(path, MAXPATHLEN, "%s%s", nbp->ndb_cachedir, cp->nc_key);

	if((fp = fopen(path, "w")) == NULL){
		perror(path);
		return -1;
	}
		
	itr = iterator_first(cp->nc_value);
	for(; !iterator_is_end(itr); itr = iterator_next(itr))
		fprintf(fp, "%s\n", iterator_data(itr));
	fclose(fp);
	return 0;
}

/* read a disk cache onto memory */
static struct cacheent *
cache_read(struct ndtpbook *nbp,  const char *key)
{
	struct cacheent *ret = NULL;
	FILE *fp;
	struct stat finfo;
	char path[MAXPATHLEN];
	char buf[1024];

	snprintf(path, MAXPATHLEN, "%s%s", nbp->ndb_cachedir, key);
	
	if(stat(path, &finfo) < 0 || (fp = fopen(path, "r")) == NULL)
		return NULL;

	else if(finfo.st_size == 0){
		unlink(path);
		fclose(fp);
		return NULL;
	}

	ret = cacheent_new(key, strlist_new(4, 4, xpdic_malloc, xpdic_free));
	while(fgets(buf, 1024, fp) != NULL){
		if(buf[strlen(buf)-1] == '\n')
			buf[strlen(buf)-1] = '\0';
		iterator_append(ret->nc_value, (const char *)xp_strdup(buf));
	}
	fclose(fp);
	return ret;
}

static int 		ndtpbook_open(struct dictbook *);
static int 		ndtpbook_close(struct dictbook *);
static void 		ndtpbook_free_(struct dictbook *);
static int 		ndtpbook_reclaim(struct dictbook *);
static strlist 		ndtpbook_lookup(struct dictbook *, const char *word);
static strlist 		ndtpbook_getentry(struct dictbook *, const char *word);
static struct bitmap 	*ndtpbook_getbitmap(struct dictbook *, const char *word, 
					    int size);
static const char  	*ndtpbook_strerr(struct dictbook *);

struct ndtpbook *
ndtpbook_new(struct ndtpserver *srvp, const char *name, int idx, const char *id)
{
	int  nfs;
	char buf[128];
	char path[MAXPATHLEN];
	struct property *prp;
	struct ndtpbook *ret = MALLOC("ndtpbook", sizeof(struct ndtpbook));

	ret->ndb_server = srvp;
	ret->ndb_name = (const char *)xp_strdup(name);
	ret->ndb_index = idx;
	ret->ndb_id = (const char *)xp_strdup(id);
	ret->ndb_errno = 0;

	/* make global id */
	snprintf(buf, 128, "%s@%s", id, srvp->ns_id);
	ret->ndb_glid = (const char *)xp_strdup(buf);

	ret->ndb_open = ndtpbook_open;
	ret->ndb_close = ndtpbook_close;
	ret->ndb_free = ndtpbook_free_;
	ret->ndb_reclaim = ndtpbook_reclaim;
	ret->ndb_lookup = ndtpbook_lookup;
	ret->ndb_getentry = ndtpbook_getentry;
	ret->ndb_getbitmap = ndtpbook_getbitmap;
	ret->ndb_strerr = ndtpbook_strerr;

	if(ndtp_setbook_idx(srvp, idx) != NDTP_SUCCESS){
		fprintf(stderr, "ndtp_setbook_idx() failed...\n");
		abort(); /* must be success */
	}
	nfs = ndtp_get_fontsize(srvp, ret->ndb_fontsizelist, MAX_NFONTSIZE);
	ret->ndb_nfontsize = (nfs < MAX_NFONTSIZE? nfs: MAX_NFONTSIZE);

	/* cache */
	ret->ndb_cache = ndtpcache_new(13, ndtpcache_hash, ndtpcache_cmp, 
				       xpdic_malloc, xpdic_free);

	/* cache path */
	ret->ndb_cachedir = NULL;

	/* check if diskcache is enabled or not */

	/* properties are put by xresource */
	prp = property_get("ndtpcache");

	if(prp && prp->prty_type == PRTY_TYPE_INTEGER &&
	   prty_integer_val(prp) != 0){
		prp = property_get("ndtpcachedir");

		if(prp && prp->prty_type == PRTY_TYPE_STRING)
			snprintf(path, MAXPATHLEN, "%s/%s/%s/", 
				 prty_string_val(prp), srvp->ns_id, id);
		else 
			snprintf(path, MAXPATHLEN, "%s/%s/%s/", 
				 "/var/tmp/.xpopdic", srvp->ns_id, id);

		/* check directory */
		if(cachedir_create(path) < 0){
			fprintf(stderr, "can't create cache dir: %s\n", path);
			fprintf(stderr, "\t--> disable disk cache\n");
		}
		else
			ret->ndb_cachedir = xp_strdup(path);
	}
	return ret;
}

static int
ndtpbook_open(struct dictbook *_dbp)
{
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;
	return ndtpserver_connect(dbp->ndb_server);
}

static int
ndtpbook_close(struct dictbook *_dbp)
{
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;
	ndtpserver_disconnect(dbp->ndb_server);
	return 1;
}

/* this function is for member of ndtpbook structure */
static void 
ndtpbook_free_(struct dictbook *_dbp)
{
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;
	booklist_iterator itr;
	if(dbp->ndb_cachedir)
		xp_strfree(dbp->ndb_cachedir);
	ndtpbook_reclaim(_dbp);
	aggregator_delete_by((ndtpcache)dbp->ndb_cache, cacheent_delete);
	xp_strfree(dbp->ndb_glid);
	xp_strfree(dbp->ndb_id);
	xp_strfree(dbp->ndb_name);

	/* find this book in book list contained by ndtpserver */
	itr = iterator_first(dbp->ndb_server->ns_books);
	for(; !iterator_is_end(itr); itr = iterator_next(itr))
		if(_dbp == iterator_data(itr)){
			iterator_erase(itr);
			break;
		}
	FREE("ndtpbook", dbp);
}

/* this function is called by ndtpserver */
void 
ndtpbook_free(struct dictbook *_dbp)
{
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;
	if(dbp->ndb_cachedir)
		xp_strfree(dbp->ndb_cachedir);
	ndtpbook_reclaim(_dbp);
	aggregator_delete_by((ndtpcache)dbp->ndb_cache, cacheent_delete);
	xp_strfree(dbp->ndb_glid);
	xp_strfree(dbp->ndb_id);
	xp_strfree(dbp->ndb_name);
	FREE("ndtpbook", dbp);
}

static int
ndtpbook_reclaim(struct dictbook *_dbp)
{
	ndtpcache_iterator itr;
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;

	if(dbp->ndb_cachedir){
		itr = iterator_first((ndtpcache)dbp->ndb_cache);

		for(; !iterator_is_end(itr); itr = iterator_next(itr)){
			if(cache_write(dbp, iterator_data(itr)->second) < 0){
				fprintf(stderr, "cache_write(%s): failed. disalbe disk cache.\n", dbp->ndb_cachedir);
				xp_strfree(dbp->ndb_cachedir);
				dbp->ndb_cachedir = NULL;
				break;
			}
		}
	}
	aggregator_clear_by((ndtpcache)dbp->ndb_cache, cacheent_delete);
	return 1;
}

static ndtpcache_iterator 
ndtpbook_cache_lookup(struct ndtpbook *dbp, const char *word)
{
	ndtpcache_iterator ret;
	struct cacheent *cep;
	ret = iterator_find((ndtpcache)dbp->ndb_cache, word);

	if(iterator_is_end(ret)){
		cep = cache_read(dbp, word);
		if(cep){
			iterator_insert_pair((ndtpcache)dbp->ndb_cache, 
					     cep->nc_key, cep);
			ret = iterator_find((ndtpcache)dbp->ndb_cache, word);
		}
	}
	return ret;
}

struct sbargs {
	strlist 	slst;
	int		nline;
};

const char *
sline_nextline(struct sbargs *sbp, char *p, int len)
{
	strlist_iterator itr = iterator_at(sbp->slst, sbp->nline);

	if(iterator_count(sbp->slst) == sbp->nline ||
	   iterator_is_end(itr))
		return NULL;

	strncpy(p, iterator_data(itr), len);
	sbp->nline++;
	return p;
}

static void 
ndtpbook_convert_bitmap(struct cacheent *p)
{
	struct sbargs sbargs;

	if(p->nc_bmp)
		return;		/* already converted */

	sbargs.slst = p->nc_value;
	sbargs.nline = 0;
	p->nc_bmp = bitmap_parse((char *(*)(void *, char *, int))sline_nextline,
				 &sbargs);
}


static int 
ndtpbook_start_transaction(struct ndtpbook *dbp)
{
	int ret;

	ret = ndtp_setbook_idx(dbp->ndb_server, dbp->ndb_index);
	if(ret != NDTP_SUCCESS)
		return 0;	/* fail */

	/* set one of fontsize */
	ret = ndtp_set_fontsize(dbp->ndb_server, dbp->ndb_fontsizelist[0]);
	
	if(ret != NDTP_SUCCESS)
		return 0;	/* fail */
	return 1;
}

static strlist 	
ndtpbook_lookup(struct dictbook *_dbp, const char *word)
{
	strlist 		slp;
	struct cacheent 	*ent = NULL;
	struct ndtpbook *dbp = (struct ndtpbook *)_dbp;
	ndtpcache_iterator 	itr = ndtpbook_cache_lookup(dbp, word);

	dbp->ndb_errno = 0;

	/* cache hit! */
	if(!iterator_is_end(itr))
		return iterator_data(itr)->second->nc_value;

	if(!ndtpbook_start_transaction(dbp)){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	slp = ndtp_lookup(dbp->ndb_server, word);
	if(!slp){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	/* create new cache entry */
	ent = cacheent_new(word, slp);
	iterator_insert_pair((ndtpcache)dbp->ndb_cache, ent->nc_key, ent);
	return slp;
}

static strlist 	
ndtpbook_getentry(struct dictbook *_dbp, const char *word)
{
	strlist 		slp;
	struct cacheent 	*ent = NULL;
	struct ndtpbook 	*dbp = (struct ndtpbook *)_dbp;
	ndtpcache_iterator 	itr = ndtpbook_cache_lookup(dbp, word);

	dbp->ndb_errno = 0;

	/* cache hit! */
	if(!iterator_is_end(itr))
		return iterator_data(itr)->second->nc_value;

	if(!ndtpbook_start_transaction(dbp)){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	slp = ndtp_getentry(dbp->ndb_server, word);
	if(!slp){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	/* create new cache entry */
	ent = cacheent_new(word, slp);
	iterator_insert_pair((ndtpcache)dbp->ndb_cache, ent->nc_key, ent);
	return slp;
}

static struct bitmap *
ndtpbook_getbitmap(struct dictbook *_dbp, const char *fname, int index)
{
	strlist 		slp;
	struct cacheent 	*ent = NULL;
	char 			fontkey[128];
	ndtpcache_iterator 	itr;
	struct ndtpbook 	*dbp = (struct ndtpbook *)_dbp;

	dbp->ndb_errno = 0;
	snprintf(fontkey, 128, "%s%d", fname, dbp->ndb_fontsizelist[index]);

	itr = ndtpbook_cache_lookup(dbp, fontkey);

	/* cache hit! */
	if(!iterator_is_end(itr)){
		ndtpbook_convert_bitmap(iterator_data(itr)->second);
		return iterator_data(itr)->second->nc_bmp;
	}

	if(!ndtpbook_start_transaction(dbp)){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	slp = ndtp_getfont(dbp->ndb_server, fname);
	if(!slp){
		if(ndtp_errno != 0)
			dbp->ndb_errno = ndtp_errno;
		return NULL;
	}

	/* create new cache entry */
	ent = cacheent_new(fontkey, slp);
	iterator_insert_pair((ndtpcache)dbp->ndb_cache, ent->nc_key, ent);
	ndtpbook_convert_bitmap(ent);
	return ent->nc_bmp;
}

static const char  *
ndtpbook_strerr(struct dictbook *bp)
{
	return ndtp_strerr(((struct ndtpbook *)bp)->ndb_errno);
}

