/*
 *			xobject.h
 * $Id: xobject.h,v 1.21 2004/02/12 00:45:58 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _XP_XOBJECT_H_
#define _XP_XOBJECT_H_

/* 
 *		xobject class hierarchy 
 *
 *	xobject*('*' means not stub)
 *	   |	+ xobj_type(xobject *);
 *	   |	+ xobj_is_enabled(xobject *);
 *	   |	+ xobj_enable(xobject *);
 *	   |	+ xobj_disable(xobject *);
 *	   |	+ xobj_is_primitive(xobject *);
 *	   |	+ xobj_is_composer(xobject *);
 *	   |	+ xobj_is_actor(xobject *);
 *	   |	+ xobj_width(xobject *);
 *	   |	+ xobj_height(xobject *);
 *	   |	+ xobj_incref(xobject *); : increment reference 
 *	   |	+ xobj_decref(xobject *); : decrement reference
 *	   |	+ xobj_free(xobject *);	
 *	   |	+ xobj_dump(xobject *, FILE *, int);
 *	   |	+ xobj_draw(xobject * obj, int x, int y, xwindow *,
 *	   | 		    int rx, int ry, int rw, int rh, int flag);	
 *	   |	   o (x, y) is the position of object 'obj'.
 *	   |	   o (dx, dy)-rw*rh is a rectangle to be drawn int the object. 
 *	   |
 * 	   +--- xprimitive*
 *	   |	     |   + xpmt_attr(xprimitive *);
 *	   |	     |   + xpmt_set_attr(xprimitive *, xattr *);
 *	   |	     |
 *	   |	     |	  each data structure is defined in xprimitive.h
 *	   |	     +----- xexpander
 *	   |	     +----- xspace
 *	   |	     +----- xpoint
 *	   |	     +----- xline
 *	   |	     +----- xrectangle
 *	   |	     +----- xarc
 *	   |	     +----- xpolygon
 *	   |	     +----- xstring
 *	   |	     +----- xwstring
 *	   |	     +----- xbitmap
 *	   |	     +----- xeditstr
 *	   |
 * 	   +--- xcomposer*
 *		     |   + xcmp_compose(xcomposer *);
 *		     |   + xcmp_relocate(xcomposer *, int x, int y);
 *		     |   + xcmp_find(xcomposer *, int x, int y, int sig, xpoint*);
 *		     |   + xcmp_wvariable(xcomposer *);
 *		     |   + xcmp_hvariable(xcomposer *);
 *		     |	   - variable width or variable height is fixed 
 *		     |     after xcmp_compose(children).
 *		     |
 *	   	     +--- xposmark
 *		     | 	  o Position mark 
 *		     | 	  o This mark caches an absolute position in the 
 *		     |	  window for an object contained by this mark.
 *		     |	  o No object other than xposmark must not cache 
 *		     |	  an absolute position.
 *		     |	  o xcmp_relocate() updates only xposmark.
 *		     |
 *	   	     +--- xactor*
 *		     |    o xactor reacts against events. 
 *		     |		+--- xclickable
 *		     |		+--- xascroll: can be moved by singal
 *		     |		+--- xkeyact: keybord event actor
 *		     |		+--- xranyact: versatile purpose actor
 *		     |		+--- xproxy: handles any signals and re-sends
 *		     |
 *	   	     +--- xcontainer*
 *		     |		+--- xvlayout
 *	   	     |		+--- xhlayout
 *	   	     |		+--- xcomposite
 *		     |
 *	   	     +--- xcompstpair(pair composition)
 *	   	     +--- xselector
 *		     +--- xscroll
 *		     +--- xframe
 *
 */

#define XOBJTYPE_UNKNOWN	0x0
#define XOBJTYPE_ANY		XOBJTYPE_UNKNOWN

/* primitive */
#define XOBJTYPE_EXPANDER	0x1
#define XOBJTYPE_SPACE		0x2
#define XOBJTYPE_POINT		0x3
#define XOBJTYPE_LINE		0x4
#define XOBJTYPE_RECTANGLE	0x5
#define XOBJTYPE_ARC		0x6
#define XOBJTYPE_POLYGON	0x7
#define XOBJTYPE_STRING		0x8
#define XOBJTYPE_WSTRING	0x9
#define XOBJTYPE_BITMAP		0xa
#define XOBJTYPE_EDITSTR	0xb

/* composer */
#define XOBJTYPE_COMPOSER	0x40
#define XOBJTYPE_POSMARK	0x41
#define XOBJTYPE_ACTOR		0x42
#define XOBJTYPE_VLAYOUT	0x43
#define XOBJTYPE_HLAYOUT	0x44
#define XOBJTYPE_COMPOSITE	0x45
#define XOBJTYPE_CMPSTPAIR	0x46
#define XOBJTYPE_SELECTOR	0x47
#define XOBJTYPE_SCROLL		0x48	
#define XOBJTYPE_FRAME		0x49

/* mask for user-defined object type */
#define XOBJTYPE_USER		0x20

/*=============  X Base Object  =============*/
/* function type for xobject */
struct xobject;
typedef void (*x_draw_t)(struct xobject *, int x, int y, 
			 xwindow *, int rx, int ry,
			 unsigned int rw, unsigned int rh, int flag);
typedef void (*x_free_t)(struct xobject *);
typedef void (*x_dump_t)(struct xobject *, FILE *, int offset);

typedef struct xobject {
	unsigned char	x_type;
	unsigned char	x_flags;	/* flags: XOBJFLAG_XXXX */
	short		x_refcnt;	/* reference count */	
	short		x_width, x_height;
	x_draw_t	x_draw;		/* draw function ptr */
	x_free_t	x_free;		/* free function ptr */
	x_dump_t	x_dump;		/* dump function ptr */
} xobject;


#define XOBJFLAG_ENABLE		0x1	/* object is enabled */
#define XOBJFLAG_ACTIVE		0x2	/* the actor is active */
#define XOBJFLAG_WVARIABLE	0x4	/* width is fixed after composing */
#define XOBJFLAG_HVARIABLE	0x8	/* height is fixed after composing */
#define XOBJFLAG_ALWACTIVE	0x10	/* the object is always active */

#define xobjcast(x)		((struct xobject *)x)
#define xobj_type(x)		xobjcast(x)->x_type
#define xobj_flags(x)		xobjcast(x)->x_flags
#define xobj_width(x)		xobjcast(x)->x_width
#define xobj_height(x)		xobjcast(x)->x_height

#define xobj_inside(o, x, y, x1, y1) \
(x <= x1 && y <= y1 && x1 < x + xobj_width(o) && y1 < y + xobj_height(o))

	/* type check macro */
#define xobj_is_primitive(x)	((xobj_type(x) & XOBJTYPE_COMPOSER) == 0)
#define xobj_is_composer(x)	((xobj_type(x) & XOBJTYPE_COMPOSER) != 0)
#define xobj_is_actor(x)	(xobj_type(x) == XOBJTYPE_ACTOR)

	/* flag checke macro */
#define xobj_is_enabled(x)	(xobj_flags(x) & XOBJFLAG_ENABLE)
#define xobj_enable(x)		xobj_flags(x) |= XOBJFLAG_ENABLE
#define xobj_disable(x) 	xobj_flags(x) &= ~XOBJFLAG_ENABLE
#define xobj_is_active(x)	(xobj_flags(x) & XOBJFLAG_ACTIVE)
#define xobj_activate(x)	xobj_flags(x) |= XOBJFLAG_ACTIVE
#define xobj_deactivate(x) 	xobj_flags(x) &= ~XOBJFLAG_ACTIVE

	/* reference count macro */
#define xobj_incref(x)	++(xobjcast(x)->x_refcnt)
#define xobj_decref(x)	--(xobjcast(x)->x_refcnt)

	/* draw the region on indicated window */
#define xobj_draw_rect_(o,x,y,z,rx,ry,rw,rh,f,clear) 			\
do {									\
	/* printf("%s: draw(%d,%d,%d,%d): (%d,%d,%d,%d)\n", xobj_typestr(o),x,y,xobj_width(o),\
	xobj_height(o),rx,ry,rw,rh);	*/				\
	if(xobj_is_enabled(o)){	/* if it is enabled */			\
		if(clear){						\
			xobjcast(o)->x_draw(xobjcast(o),x,y,z,rx,ry,rw,rh,\
				f | XOBJDRAW_REDRAW);			\
		}							\
		else							\
			xobjcast(o)->x_draw(xobjcast(o),x,y,z,rx,ry,rw,rh,f);\
	}								\
} while(0)

#define xobj_draw_rect(o,x,y,z,rx,ry,rw,rh,f) 				\
		xobj_draw_rect_(o,x,y,z,rx,ry,rw,rh,f,0) 
#define xobj_redraw_rect(o,x,y,z,rx,ry,rw,rh,f) 			\
		xobj_draw_rect_(o,x,y,z,rx,ry,rw,rh,f,1) 

	/* draw the object on indicated window */
#define xobj_draw(o,x,y,z,f) 						\
	xobj_draw_rect(o,x,y,z,x,y,xobj_width(o),xobj_height(o),f)
#define xobj_redraw(o,x,y,z,f) 						\
	xobj_redraw_rect(o,x,y,z,x,y,xobj_width(o),xobj_height(o),f)

	/* draw object on the top of object tree */
#define xobj_draw_object(r,o,x,y,w,f)					\
	xobj_draw_rect(r,0,0,w,x,y,xobj_width(o),xobj_height(o),f)

#define xobj_redraw_object(r,o,x,y,w,f)					\
	xobj_redraw_rect(r,0,0,w,x,y,xobj_width(o),xobj_height(o),f)
		
#define xobj_free(x)							\
	do { 								\
		xobj_decref(x);						\
		if(xobjcast(x)->x_refcnt <= 0)				\
			xobjcast(x)->x_free(xobjcast(x));		\
	} while(0)
#define xobj_dump(x, y, z)   xobjcast(x)->x_dump(xobjcast(x), y, z)


/* flag to draw objects with inactive color */
#define XOBJDRAW_INACTIVE	0x1
#define XOBJDRAW_REDRAW		0x2	/* redraw for frame/scroll */

/* utility functions for xobj_dump() */
const char 	*xod_typestr(int);
int		xod_indent(char *bufp, int size);
#define	xobj_typestr(x)	(xod_typestr(xobj_type(x)))
int	xobj_info(struct xobject *, char *bufp, int size);


/*=============  X Primitive Object  =============*/
typedef struct xprimitive {
	struct xobject xp_base;
	xattr	       *xp_attr;
} xprimitive;
#define xpmtcast(x)	((struct xprimitive *)x)
#define xpmt_attr(x)  	xpmtcast(x)->xp_attr
#define xpmt_set_attr(x, a)  (xpmtcast(x)->xp_attr = a)

	
/*=============  X Composer Object  =============*/
typedef struct xcomposer {
	struct xobject xc_base;
	void   	       (*xc_compose)(struct xcomposer *);
	void   	       (*xc_relocate)(struct xcomposer *, int x, int y);
	struct xobject *(*xc_find)(struct xcomposer *, int x, int y, int flags,
				   xpoint *ret_offset);
} xcomposer;

	/* composition rule definition for alignment */
#define XCRULE_CENTER		0x0
#define XCVRULE_CENTER		0x0
#define XCHRULE_CENTER		0x00
#define XCRULE_TOP		0x1	
#define XCRULE_BOTTOM		0x2
#define XCRULE_LEFT		0x10
#define XCRULE_RIGHT		0x20

/* variable size which is fixed when composing */
#define XC_VARIABLE		-1	
#define xcmpcast(x)	 ((struct xcomposer *)x)
#define xcmp_compose(x)  xcmpcast(x)->xc_compose(xcmpcast(x))
#define xcmp_relocate(o, x, y) xcmpcast(o)->xc_relocate(xcmpcast(o), x, y)
#define xcmp_find(o, x, y, m, p) 				\
	    (xobj_is_enabled(o)? xcmpcast(o)->xc_find(xcmpcast(o), x, y, m, p): NULL)

#define xcmp_wvariable(x)  ((xobj_flags(x) & XOBJFLAG_WVARIABLE) != 0)
#define xcmp_hvariable(x)  ((xobj_flags(x) & XOBJFLAG_HVARIABLE) != 0)


/* find flags definitions */
/* XXXX: actually only xactor/xposmark can be indicated as type currently */
#define XCMPFIND_FIRSTMATCH	0x00000000	/*  return object matching indicated 
						   conditions  first */
#define XCMPFIND_LASTMATCH	0x80000000	/*  return object matching indicated 
						   conditions last */
#define xcmpfind_build_flags(f, type, subtype)   (f | (type << 24) | subtype)

#define xcmpfind_type(f)   		(0x7f & ((f) >> 24))
#define xcmpfind_typeany(f)   		(xcmpfind_type(f) == 0)
#define xcmpfind_subtype(f)   		(0x00ffffff & (f))
#define xcmpfind_is_firstmatch(f)	(!xcmpfind_is_lastmatch(f))
#define xcmpfind_is_lastmatch(f)	(XCMPFIND_LASTMATCH & (f))


/*=============  X Container Object  =============*/
/* Data format contained by X composers */
typedef struct xoposition { 
	unsigned char		xp_rule; /* XCVRULE_XXX, XCHRULE_XXX */
	unsigned char		_xp_pad1;
	unsigned short		_xp_pad2;
	struct xpoint 		xp_offset;	
	struct xobject		*xp_object;
} xoposition;

typedef struct xcontainer {
	struct xcomposer xc_base;
	void		*xc_list;   /* xoplist */
} xcontainer;

#define xcontcast(x)	((struct xcontainer *)x)
#define xcont_list(x) 	 (xoplist)xcontcast(x)->xc_list
void    xcont_add(struct xobject *, struct xobject *, unsigned int rule);
int     xcont_count(struct xobject *);
struct xoposition *xcont_at(struct xobject *, int idx);
int     xcont_theight(struct xobject *); /* total height */
int     xcont_twidth(struct xobject *);  /* total width */

/*=============  X Position Mark Object  =============*/
/* 
 *  X Position mark object. 
 *	- this object is used for keep an absolute position for 
 * 	a special object.
 *	- this object cannot be shared by other composers.
 *	- useful to send a signal to other objects
 */
typedef struct xposmark {
	struct xcomposer xp_base;
	xobject		*xp_obj;
	xpoint		 xp_pos;
	void		*xp_data;	/* user data */
} xposmark;

#define xpm_obj(p)	((struct xposmark *)p)->xp_obj
#define xpm_pos(p)	((struct xposmark *)p)->xp_pos
#define xpm_x(p)	((struct xposmark *)p)->xp_pos.x
#define xpm_y(p)	((struct xposmark *)p)->xp_pos.y
#define xpm_data(p)	((struct xposmark *)p)->xp_data


/*=============  X Object Scroller  =============*/
typedef struct xscroll {
	struct xcomposer xs_base;
} xscroll;
#define xscrlcast(s) 	((struct xscroll *)s)
#define xsc_scroll(s,x,y,w,r,x1,y1) 	xscrl_scroll(xscrlcast(s),x,y,w,r,x1,y1)
#define xsc_reset(s,x,y,w,r) 		xscrl_reset(xscrlcast(s),x,y,w,r)
#define xsc_move(s,x,y,mx,my) 		xscrl_move(xscrlcast(s),x,y,mx,my)
void   xscrl_scroll(struct xscroll *, int, int, xwindow *, struct xobject *, int, int);
void   xscrl_reset(struct xscroll *, int, int, xwindow *, struct xobject *);
void   xscrl_move(struct xscroll *, int, int, int, int);


/*=============  X Composite Pair =============*/
/*  xcmpstpair composes two prmitive objects to perform
 * special effects such as underline.
 *  xcmpstpair can be applied for xcomposer's but 
 * in such cases, xcomposite should be used.
 */
typedef struct xcmpstpair {
	struct xcomposer 	xcp_base;
	struct xobject		*xcp_first;
	struct xobject		*xcp_second;
	struct xpoint		xcp_firstoft;  /* offset */
	struct xpoint		xcp_secondoft;  /* offset */
	unsigned char		xcp_rule;
	unsigned char		_pad1;
	unsigned short		_pad2;
} xcmpstpair;


/*=============  X Object Selector =============*/
typedef struct xselector {
	struct xcomposer xs_base;
	unsigned char    xs_idx;	 /* current object */
	unsigned char    xs_nobjs;       /* number of object */
	short	   	 __xs_pad;       /* padding */
	struct xobject	**xs_objs; 
} xselector;

#define xselcast(x)	((struct xselector *)x)
/*  
 *  put a new object into xselector. if there is already an object, 
 * it will be xobj_free'd.
 */
void	xsel_set(struct xobject *, unsigned char idx, struct xobject *);

/* 
 *  xsel_select() changes the drawn object to the object indicated by 
 * the 3rd artument. call xobj_redraw() after calling this function.
 */
void	xsel_select(struct xobject *, unsigned char idx);


/*===========  Function Declarations  ===========*/
/* X primitive objects */
struct xobject *xexpander_new(void);
struct xobject *xspace_new(int width, int height);
struct xobject *xline_new(int x1, int y1, int x2, int y2, xattr *);
struct xobject *xrect_new(int w, int h, int fill, xattr *);
struct xobject *xarc_new(int w, int h, int a1, int a2, int fill, xattr *);
struct xobject *xbitmap_new(struct bitmap *, struct xattr *);
struct xobject *xstring_new(const char *str, int nbytes, struct xattr *, int bg);
struct xobject *xwstring_new(const char *str, int nchars, struct xattr *, int bg);

/* these two functions don't allocate memory for the indicated string. */
struct xobject *xstring_na_new(const char *str, int nbytes, struct xattr *, int bg);
struct xobject *xwstring_na_new(const char *str, int nchars, struct xattr *, int bg);

/* to draw underline for the object */
#define underline_new(x, a)	\
	xline_new(0, xobj_height(x)-1, xobj_width(x)-1, xobj_height(x)-1, a)

/* X container objects and a related object */
struct xobject *xposmark_new(struct xobject *target);
struct xobject *xvlayout_new(unsigned int width, unsigned int height, 
				unsigned int mergin, unsigned int layout_rule);
struct xobject *xhlayout_new(unsigned int width, unsigned int height, 
				unsigned int mergin, unsigned int layout_rule);
struct xobject *xcomposite_new(unsigned int width, unsigned int height);
struct xobject *xscroll_new(struct xobject *, unsigned int w, unsigned int h);
struct xobject *xscroll_new_bg(struct xobject *, unsigned int w, unsigned int h, 
			       xattr *);
struct xobject *xframe_new(struct xobject *, unsigned int w, unsigned int h);
struct xobject *xframe_new_bg(struct xobject *, unsigned int w, unsigned int h,
			      xattr *);
struct xobject *xcmpstpair_new(struct xobject *, struct xobject *, int);
struct xobject *xselector_new(unsigned char nobjs, unsigned int width, 
			      unsigned int height);

#endif /* _XP_XOBJECT_H_ */

