/*
 *			xproxy.c
 * $Id: xproxy.c,v 1.6 2004/01/24 10:44:31 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define _XACTOR_INTERNAL
#include <stdio.h>
#include "memory.h"
#include "bitmap.h"
#include "xwindow.h"
#include "xobject.h"
#include "xsignal.h"
#include "xactor.h"
#include "resource.h"
#include "xsigset.h"

/* xembed = object + position */
struct xembed {
	struct xactor 	*o;
	xpoint		pos;
};
#include "xemblist.c"

/* internal object type for xproxy */
struct _xproxy {
        struct xproxy   	_base;
	struct xaction  	*any;        /* action for any signal */
	xsigset            	accepts;    /* acceptable signal set */
	xemblist	 	entrstack;  /* entering object stack */
	struct xembed		ecur;	    /* current entered object */
	struct xembed		bpress;	    /* button pressed object */
	struct xembed		kpress;		/* key pressed object */
	int			lastx, lasty; /* start point of button motion */
};

/* push/pop xembedobject */
void push_enter(struct _xproxy *c)
{
	struct xembed *e = MALLOC("xembed", sizeof(struct xembed)); 

	if(c->ecur.o == NULL)
		abort();
	*e = c->ecur;
	iterator_append(c->entrstack, e);
	c->ecur.o = NULL;
}

void pop_enter(struct _xproxy *c)
{
	xemblist_iterator itr = iterator_last(c->entrstack);
	c->ecur.o = NULL;

	/* empty */
	if(iterator_count(c->entrstack) == 0)
		return;

	c->ecur = *iterator_data(itr);
	FREE("xembed", iterator_data(itr));
	iterator_erase(itr);
}

/* free functions for aggregator_delete_by() */
static void xpr_efree(struct xembed *r) { FREE("xembed", r); }

/* functions for xsigset */
int  xsigcmp(int x, int y) { return x-y; }
unsigned int xsighash(int x) { return x; }

/* clear cache */
static void 
xpr_reset_cache(struct _xproxy *c, struct xobject *root, xwindow *xwp)
{
	/* check the position is still inside of the entered objects */
	while(c->ecur.o != NULL){
		struct xsigarg sigarg = {c->ecur.pos.x, c->ecur.pos.y };
		xact_send_signal(c->ecur.o, c->ecur.pos, XSIG_LEAVE, &sigarg, 
				 root, xwp);
		pop_enter(c);
	}
	c->ecur.o = c->bpress.o = c->kpress.o = NULL;
}

static void
xpr_free(struct xobject *o)
{
	struct _xproxy *c = (struct _xproxy *)o;
	aggregator_delete_by(c->entrstack, xpr_efree);
	aggregator_delete(c->accepts);
	xobj_free(((struct xactor *)c)->xa_object);
	if(c->any)
		xaction_delete(c->any);
	FREE("xproxy", c);
}

#define actdump(x, id, f, i, idt)  do {					\
	if(!x)								\
		fprintf(fp, "%s %s: not prepaired.\n", idt, id); 	\
	else if(!(x)->xac_object)					\
		fprintf(fp, "%s %s: no object to be drawn.\n", idt, id); \
	else {								\
		fprintf(fp, "%s %s:\n", idt, id); 			\
		xobj_dump((x)->xac_object, fp, i + 1);		\
	}								\
} while(0)

static void
xpr_dump(struct xobject *o, FILE *fp, int off)
{
	struct _xproxy *c = (struct _xproxy *)o;
	struct xactor *r = (struct xactor *)o;
	char idt[1024], buf[1024], info[1024], *p = idt;

	xod_indent(idt, off);
	xobj_info(o, info, 1024);
	
	snprintf(buf, 1024, "%s%s<Signal Proxy>: %s{\n", 
		 idt, xobj_typestr(o), info);
	fprintf(fp, "%s", buf);

	while(*p)
		*p++ = ' ';
	actdump(c->any, "Any", fp, off, idt); 
	fprintf(fp, "%s Main:\n", idt); 
	xobj_dump(r->xa_object, fp, off + 1);
	fprintf(fp, "%s}\n", idt);
}

static void
xpr_compose(struct xcomposer *o)
{
	struct xactor *r = (struct xactor *)o;
	struct _xproxy *c = (struct _xproxy *)o;

	if(xobj_is_composer(r->xa_object))
		xcmp_compose(r->xa_object);
	if(xcmp_wvariable(o))
		xobj_width(o) = xobj_width(r->xa_object);
	if(xcmp_hvariable(o))
		xobj_height(o) = xobj_height(r->xa_object);

	xact_compose(c->any);
}

static int
xpr_acceptable(struct xactor *r, xsignal s)
{
	struct _xproxy *c = (struct _xproxy *)r;
	xsigset_iterator itr = iterator_find(c->accepts, s);

	/* acceptable */
	if(!iterator_is_end(itr))
		return 1;

	/* not acceptable, check XSIG_ANY */
	itr = iterator_find(c->accepts, XSIG_ANY);
	if(!iterator_is_end(itr))
		return 1;

	/* can't accept */
	return 0;
}

static struct xaction *
xpr_getact(struct xactor *r, xsignal s)
{
	struct _xproxy *c = (struct _xproxy *)r;
	return c->any;
}

static int
xpr_register(struct xactor *r, xsignal s, xaction_t a, 
	     void *p, struct xobject *o)
{
	struct _xproxy *c = (struct _xproxy *)r;
	struct xaction **dst = NULL;

	switch(xsignal_type(s)){
	case XSIG_ANY:
		dst = &c->any;
		break;
	}

	/* unacceptable signal */
	if(!dst)
		return 0;  /* fail */

	/* delete old action */
	if(*dst)
		xaction_delete(*dst);

	*dst = xaction_new(o, a, p);
	return 1;	/* success */
}


static void
xpr_receive(struct xactor *act, xpoint pos, xsignal sig, struct xsigarg *sigarg, 
	    struct xobject *root, xwindow *w);

struct xproxy *
xproxy_new(struct xcomposer *o)
{
	struct _xproxy *c = MALLOC("xproxy", sizeof(struct _xproxy));
	struct xactor *ret = (struct xactor *)c;

	xactor_init(ret, (struct xobject *)o, xactor_draw, xpr_free, xpr_dump, 
		    xpr_compose, xactor_find, xactor_relocate);

	ret->xa_acceptable = xpr_acceptable;
	ret->xa_getaction = xpr_getact;
	ret->xa_register = xpr_register;
	ret->xa_receive = xpr_receive;
	c->any = NULL;
	c->accepts = xsigset_new(3, xsighash, xsigcmp, stk_malloc, stk_free);
	c->entrstack = xemblist_new(4, 4, stk_malloc, stk_free);
	c->ecur.o = NULL;
	c->bpress.o = NULL;
	c->kpress.o = NULL;
	c->lastx = c->lasty = 0;
	xproxy_accept((struct xproxy *)c, XSIG_PROXY_CACHERESET);
	return (struct xproxy *)ret;
}

void 
xproxy_accept(struct xproxy *x, xsignal s)
{
	struct _xproxy *c = (struct _xproxy *)x;
	iterator_insert(c->accepts, s);
}

/* these events must be cared */
static void btn_pressed(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig, 
			struct xsigarg *sigarg, struct xobject *root);
static void btn_released(xwindow *xwp,   struct _xproxy *c, xpoint pos, xsignal sig, 
			 struct xsigarg *sigarg, struct xobject *root);
static void btn_motion(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig, 
		       struct xsigarg *sigarg, struct xobject *root);
static void key_pressed(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig,
			struct xsigarg *sigarg, struct xobject *root);
static void key_released(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig,
			 struct xsigarg *sigarg, struct xobject *root);

static void
xpr_receive(struct xactor *act, xpoint pos, xsignal sig, struct xsigarg *sigarg, 
	    struct xobject *root, xwindow *w)
{
	struct _xproxy *c = (struct _xproxy *)act;
	int 		flags, nx = 0, ny = 0;
	xobject 	*o;
	xpoint 		op = {0, 0};

	switch(xsignal_type(sig)){
	case XSIG_BTNPRESS:
		btn_pressed(w, c, pos, sig, sigarg, root);
		break;
	case XSIG_BTNRELEASE:
		btn_released(w, c, pos, sig, sigarg, root);
		break;
	case XSIG_MOTION:
		btn_motion(w, c, pos, sig, sigarg, root);
		break;
	case XSIG_KEYPRESS:
		key_pressed(w, c, pos, sig, sigarg, root);
		break;
	case XSIG_KEYRELEASE:
		key_released(w, c, pos, sig, sigarg, root);
		break;
	case XSIG_PROXY_CACHERESET:
	case XSIG_LEAVE:
		xpr_reset_cache(c, root, w);
		break;
	default:
		/* culculate related position value */
		if(sigarg){
			/* position contained by sigarg is absolute value */
			nx = sigarg->x - pos.x;
			ny = sigarg->y - pos.y;
		}
		/* other signal is just send to inside object */
		flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, XOBJTYPE_ANY, 
					     xsignal_type(sig));

		/* find object on (nx,ny) */
		o = xcmp_find(act->xa_object, nx, ny, flags, &op);   /* XXX */

		/* adajast position to absolute value */
		op.x += pos.x;
		op.y += pos.y;

		if(o && xobj_is_actor(o)){
			xactor *a = (xactor *)o;
			xact_send_signal(a, op, sig, sigarg, root, w);
		}
	}
}

static void 
btn_pressed(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig, 
	    struct xsigarg *sigarg, struct xobject *root)
{
	int 		nx = 0, ny = 0;
	struct xobject 	*o;
	struct xactor   *act = (struct xactor *)c;
	xpoint 		op = {0, 0};
	int flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, 
					 XOBJTYPE_ACTOR, XSIG_BTNPRESS);

	/* XXXX: pressed object exists! */
	if(c->bpress.o)
		c->bpress.o = NULL;	

	/* find press-able object in the entered object */
	if(c->ecur.o){
		nx = sigarg->x - c->ecur.pos.x;
		ny = sigarg->y - c->ecur.pos.y;
		o = xcmp_find(c->ecur.o, nx, ny, flags, &op);
		op.x += c->ecur.pos.x;
		op.y += c->ecur.pos.y;
	}
	else {
		nx = sigarg->x - pos.x;
		ny = sigarg->y - pos.y;

		o = xcmp_find(act->xa_object, nx, ny, flags, &op);   
		op.x += pos.x;
		op.y += pos.y;
	}

	if(o && xobj_is_actor(o)){
		xactor *a = (xactor *)o;
		xact_send_signal(a, op, sig, sigarg, root, xwp);
		c->bpress.o = a;
		c->bpress.pos = op;
	}
	else
		c->bpress.o = NULL;
	c->lastx = sigarg->x;
	c->lasty = sigarg->y;
}

static void 
btn_released(xwindow *xwp,   struct _xproxy *c, xpoint pos, xsignal sig, 
	     struct xsigarg *sigarg, struct xobject *root)
{
        int maxx, maxy;

	/* BTNPRESS signal would be lost */
	if(c->bpress.o == NULL)
		return;		/* just ignore */

	maxx = c->bpress.pos.x + xobj_width(c->bpress.o);
	maxy = c->bpress.pos.y + xobj_height(c->bpress.o);

	/* the point is outsize? */
	if(sigarg->x < c->bpress.pos.x || sigarg->y < c->bpress.pos.y
	|| sigarg->x >= maxx || sigarg->y >= maxy)
		sig = XSIG_LEAVE;

	xact_send_signal(c->bpress.o, c->bpress.pos, sig, sigarg, root, xwp);
	c->bpress.o = NULL;
	c->lastx = c->lasty = 0;
}

static void 
btn_motion(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig, 
	   struct xsigarg *sigarg, struct xobject *root)
{
	int flags, nx = 0, ny = 0;
	struct xs_btninfo *bmp = (struct xs_btninfo *)sigarg;
	struct xactor 	  *act = (struct xactor *)c;
	struct xobject 	  *o;
	xpoint 		   op = {0, 0};


	/* no button is pressed */
	if(bmp->button == -1){

		/* current entered object is not set */	
find_again:
		if(c->ecur.o == NULL){
			/* make relational position from position of c */
			nx = sigarg->x - pos.x;
			ny = sigarg->y - pos.y;

			/* find! */
			flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, 
						     XOBJTYPE_ANY, XSIG_ENTER);
			o = xcmp_find(act->xa_object, nx, ny, flags, &op);   
			op.x += pos.x;
			op.y += pos.y;

			if(o && xobj_is_actor(o)){
				struct xactor *a = (struct xactor *)o;

				xact_send_signal(a, op, XSIG_ENTER, 
					     		sigarg, root, xwp);
				c->ecur.o = a;
				c->ecur.pos = op;
			}
			return;
		}

		/* check the position is still inside of the entered objects */
		while(c->ecur.o != NULL){
			int x = sigarg->x, y = sigarg->y;
		        int maxx = c->ecur.pos.x + xobj_width(c->ecur.o);
			int maxy = c->ecur.pos.y + xobj_height(c->ecur.o);

			/* the point is inside? */
			if(x >= c->ecur.pos.x && y >= c->ecur.pos.y 
			&& x < maxx && y < maxy)
				break;

			/* outside! */
			xact_send_signal(c->ecur.o, c->ecur.pos, XSIG_LEAVE, sigarg, 
					 root, xwp);
			pop_enter(c);
		}

		/* try to find */
		if(c->ecur.o == NULL)
			goto find_again;

		/* first, find an actor accepting XSIG_MOTION */
		flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, 
					     XOBJTYPE_ANY, XSIG_MOTION);

		/* if there is nested proxy, it will accept this signal */
		nx = sigarg->x - c->ecur.pos.x;
		ny = sigarg->y - c->ecur.pos.y;
		o = xcmp_find(c->ecur.o, nx, ny, flags, &op);   

		/* everything will be done by next actor. So we don't care anymore */
		if(o && xobj_is_actor(o)){
			struct xactor *a = (struct xactor *)o;
			op.x += c->ecur.pos.x;
			op.y += c->ecur.pos.y;
			/* everything will be done by this actor */
			xact_send_signal(a, c->ecur.pos, XSIG_MOTION, sigarg, root, xwp);
			return; 
		}


		/* find a nested actor on the current entered actor */
		flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, 
					     XOBJTYPE_ANY, XSIG_ENTER);

		/* find inside of the entered actor */
		if(!xobj_is_composer(c->ecur.o->xa_object))
			return;
		/* c->ecur.o->xa_object -> prevented to influence of actor */
		o = xcmp_find(c->ecur.o->xa_object, nx, ny, flags, &op); 

		if(o && xobj_is_actor(o)){
			struct xactor *a = (struct xactor *)o;
			op.x += c->ecur.pos.x;
			op.y += c->ecur.pos.y;

			/* push current entered actor on a stack */
			push_enter(c);
			c->ecur.o = a;
			c->ecur.pos = op;

			/* send signal to nested actor */
			xact_send_signal(a, op, XSIG_ENTER, sigarg, root, xwp);
		}
		return;
	}

	/* button is pressed */
	else {
		struct xs_bmotion bmt;

		/* button pressed signal would have been lost */
		if(c->bpress.o == NULL)
			return;		/* just ignore */

		/* this object can't accept GRABMOTION */
		if(!xact_acceptable(c->bpress.o, XSIG_GRABMOTION))
			return;

		bmt.pos.x = sigarg->x;
		bmt.pos.y = sigarg->y;
		bmt.button = bmp->button;
		bmt.mx = bmt.pos.x - c->lastx;
		bmt.my = bmt.pos.y - c->lasty;
		c->lastx = bmt.pos.x;
		c->lasty = bmt.pos.y;

		xact_send_signal(c->bpress.o, c->bpress.pos, XSIG_GRABMOTION,
				  (struct xsigarg *)&bmt, root, xwp);
	}
}


static void 
key_pressed(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig,
	    struct xsigarg *sigarg, struct xobject *root)
{
	int 		flags, nx = 0, ny = 0;
	xobject 	*o;
	xpoint 		op = {0, 0};
	struct xactor	*act = (struct xactor *)c;
	
	/* XSIG_KEYRELEASE signal would be lost */
	if(c->kpress.o != NULL)
		c->kpress.o = NULL;	/* just ignore */


	flags = xcmpfind_build_flags(XCMPFIND_FIRSTMATCH, XOBJTYPE_ANY, sig);

#if 0
	/* find object on (nx,ny) */
	/* find press-able object in the entered object */
	if(c->ecur.o){
		/* sigarg must not be NULL */
		nx = sigarg->x - c->ecur.pos.x;
		ny = sigarg->y - c->ecur.pos.y;
		o = xcmp_find(c->ecur.o, nx, ny, flags, &op);
		op.x += c->ecur.pos.x;
		op.y += c->ecur.pos.y;
	}
	else {
		/* sigarg must not be NULL */
		nx = sigarg->x - pos.x;
		ny = sigarg->y - pos.y;
		o = xcmp_find(act->xa_object, nx, ny, flags, &op);   
		op.x += pos.x;
		op.y += pos.y;
	}
#endif
	/* sigarg must not be NULL */
	nx = sigarg->x - pos.x;
	ny = sigarg->y - pos.y;
	o = xcmp_find(act->xa_object, nx, ny, flags, &op);   
	/* adajast position to absolute value */
	op.x += pos.x;
	op.y += pos.y;

	if(o && xobj_is_actor(o)){
		xactor *a = (xactor *)o;
		xact_send_signal(a, op, sig, sigarg, root, xwp);
		c->kpress.o = a;
		c->kpress.pos = op;
	}
}

static void 
key_released(xwindow *xwp, struct _xproxy *c, xpoint pos, xsignal sig,
	     struct xsigarg *sigarg, struct xobject *root)
{
	/* button was pressed */
	if(c->kpress.o != NULL){
		xact_send_signal(c->kpress.o, c->kpress.pos, sig, sigarg, root, xwp);
		c->kpress.o = NULL;
	}
}

