/*
 *			xpopdic_control.c
 * $Id: xpopdic_control.c,v 1.12 2004/02/12 00:45:58 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "debug.h"
#include "memory.h"
#include "property.h"
#include "bitmap.h"
#include "wordhist.h"
#include "dictbook.h"
#include "booklist.h"
#include "dictlink.h"
#include "xwindow.h"
#include "stk/stkbase.h"
#include "strset.h"
#include "xobjcache.h"
#include "xpopdic_resource.h"
#include "xpopdic_component.h"
#include "xpopdic_control.h"
#include "xpstatus.h"

/* function declarations for only xpcontrol_initialize() */
struct xpopdic_resource *xpresource_initialize(xrootwindow *);
void xpresource_finalize(xrootwindow *);
struct xpopdic_component *
xpcomponent_new(struct xpopdic_resource *,struct xpopdic_control *);
void xpcomponent_delete(struct xpopdic_component *);

/* static variabale for booklink error handle */
static struct xpopdic_control xpctrl;

/* static function */
static xobject *xppage_build(struct xpopdic_control *cp, struct dictnode *);
static struct dictnode *xpc_word_lookup(booklist books, const char *word, 
                           	    	struct dictnode *current);

/* functions for booklink */
static void 
xdictfree(void *p)
{ 
	xobj_free(p); 
}
static void 
xdicterr(const char *dp, const char *estr, const char *cause)
{
	int mapnow = (xpctrl.xo_cur != NULL);
	char buf[256];

	if(mapnow){
		snprintf(buf, 256, "Error: \"%s\" on %s", estr, dp);
		xpc_send_status(&xpctrl, XSIG_STAT_WARN, buf);
		xroot_flush(xpctrl.win->xw_root);
		sleep(3);
	}
	else
		fprintf(stderr, "xdicterr(): %s/%s/%s\n", dp, estr, cause);
}


struct xpopdic_control *
xpcontrol_initialize(xrootwindow *rwp, booklist books)
{
	int fixed_x = 0, fixed_y = 0;
	struct xpopdic_resource *xpr;
	/* initialize resources */
	xpr = xpctrl.xpr = xpresource_initialize(rwp);

	/* initialize dictlink */
	dictlink_init(books, xdicterr);
	dictlink_set_viewfree(xdictfree);

	/* initialize stk */
	stk_initialize(rwp, xpdic_malloc, xpdic_free);

	/* fixed position to popup */
	fixed_x = (xpr->x != -1)? xpr->x: 0;
	fixed_y = (xpr->y != -1)? xpr->y: 0;

	/* open window */
	xpctrl.win = xwindow_create(rwp, fixed_x, fixed_y, xpr->width, 
				  xpr->height, xpr->bdwidth, xpr->bgc, 
				  xpr->bdc, (xpr->shadow != 0), 1);

	/* set variables */
	xpctrl.books = books;
	xpctrl.dn_cur = NULL;
	xpctrl.xo_cur = NULL;
	xpctrl.xo_top = NULL;
	xpctrl.do_quit = 0;

	xpctrl.xpc = xpcomponent_new(xpr, &xpctrl);
	return &xpctrl;
}

void 
xpcontrol_finalize(xrootwindow *rwp, struct xpopdic_control *cp)
{
	if(cp->xo_top)
		xobj_free(cp->xo_top);
	if(cp->dn_cur)
		dl_quit(cp->books, cp->dn_cur);
	
	xpcomponent_delete(cp->xpc);	/* reuse components */
	xwindow_destroy(cp->win);	/* destroy window */
	stk_finalize(rwp);		/* stk library */
	dictlink_fin(cp->books);	/* dictionary */
	xpresource_finalize(rwp);	/* resources */
	memset(cp, 0, sizeof(*cp));
}

/* map window & draw */
void 
xpc_popup(struct xpopdic_control *cp)
{
	int curx, cury;
	int nx, ny;

	/* get current mouse position */
	xroot_getptr(cp->win->xw_root, &curx, &cury);

	/* make decision where pop window appears */
	nx = (cp->xpr->x != -1)? cp->xpr->x : curx;
	ny = (cp->xpr->y != -1)? cp->xpr->y : cury;

	/* move! */
	xw_move(cp->win, nx, ny);

	/* map and raise */
	stk_map();
	xw_raise(cp->win);
	xroot_flush(cp->win->xw_root); /* flush once */

	/* draw! */
	stk_draw();
	xroot_flush(cp->win->xw_root); /* flush again */
}

static void
xpc_set_next(struct xpopdic_control *cp, struct dictnode *next)
{
	int mapnow = (cp->xo_cur != NULL);

	/* build page */
	if(!next->dn_view)
		next->dn_view = xppage_build(cp, next);

	cp->dn_cur = next;
	cp->xo_cur = (xobject *)next->dn_view;

	stk_set(cp->win, cp->xo_cur);

	/*  send signal statusbar */ 
	xpc_send_status(cp, XSIG_STAT_NORMAL, next->dn_whist->whe_ldatestr);

	if(mapnow)
		stk_redraw();

	if(DEBUG_IS_ENABLED(DEBUG_XOBJECT))
		xobj_dump(next->dn_view, stdout, 0);
}

/* main control interfaces */
static void
xpc_move(struct xpopdic_control *cp, 
	 struct dictnode *(*f)(booklist, struct dictnode *))
{
	struct dictnode *next = (*f)(cp->books, cp->dn_cur);
	
	if(!next)
		return;
	xpc_set_next(cp, next);
}

void xpc_up(struct xpopdic_control *cp) 
{ 
	stk_set(NULL, NULL); /* reset first */
	xpc_move(cp, dl_up); 
}
void xpc_next(struct xpopdic_control *cp) { xpc_move(cp, dl_next); }
void xpc_prev(struct xpopdic_control *cp) { xpc_move(cp, dl_prev); }

void 
xpc_search(struct xpopdic_control *cp, const char *word)
{
	struct dictnode *next;
	int mapnow = (cp->xo_cur != NULL);

	if(mapnow){
		char buf[256];
		snprintf(buf, 256, "Searching \"%s\" now...", word);
		/* send signal to status bar */
		xpc_send_status(cp, XSIG_STAT_UP, buf);
	}
	
	next = xpc_word_lookup(cp->books, word, cp->dn_cur);	

	if(!next){
		if(mapnow){
			char buf[256];
			snprintf(buf, 256, "No entry for \"%s\" was found...", word);
			/* send signal to status bar */
			xpc_send_status(cp, XSIG_STAT_WARN, buf);

			xroot_flush(cp->win->xw_root);
		}
		return;
	}

	xpc_set_next(cp, next);
}

void 
xpc_go_entry(struct xpopdic_control *cp, const char *entryid)
{
	struct dictnode *next = dl_entry(cp->books, entryid, cp->dn_cur);	

	if(!next){
		/* send signal to status bar */
		return;
	}
	xpc_set_next(cp, next);
}

void 
xpc_go_at(struct xpopdic_control *cp, unsigned int i)
{
	struct dlistent *ent = dl_list_at(cp->dn_cur, i);
	
	if(!ent){
		/* send signal to status bar */
		return;
	}
	xpc_go_entry(cp, ent->dl_key);
}

void 
xpc_close(struct xpopdic_control *cp)
{
	stk_unmap();		/* close window */
	xroot_flush(cp->win->xw_root);
	stk_stop_loop();	/* stop loop */
	stk_set(NULL, NULL);	/* reset object */

	/* free current booknode tree and save ndtp cache */
	dl_quit(cp->books, cp->dn_cur);
	
	cp->dn_cur = NULL;
	cp->xo_cur = NULL;
}



void 
xpc_send_status(struct xpopdic_control *cp, xsignal sig, const char *msg)
{
	struct xps_sigarg stat;
	xobject 	  *stp = cp->xpc->xo_status;

	stat.pos.x = xpm_x(stp);
	stat.pos.y = xpm_y(stp);
	stat.msg = msg;
	stat.nchars = msg? strlen(msg): 0;

	/* send signal to status bar */
	xact_send_signal((xactor *)xpm_obj(stp), xpm_pos(stp), sig, 
			 (struct xsigarg *)&stat, cp->xo_cur, cp->win);
	xroot_flush(cp->win->xw_root);
}

static void 
xpc_btn_send(struct xpopdic_control *cp, xobject *a, xsignal sig)
{
	struct xs_btninfo sigarg = { { xpm_x(a), xpm_y(a)}, 0 };

	/* 
	 *  set redirect flag to prevent the clickable button change
	 * the cursor.
	 */
	xsignal_set_redirect(sig);

	/* send signal to status bar */
	xact_send_signal((xactor *)xpm_obj(a), xpm_pos(a), sig, 
			 (struct xsigarg *)&sigarg, cp->xo_cur, cp->win);
}

void 
xpc_press_quit(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_quit, XSIG_BTNPRESS);
}
void 
xpc_release_quit(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_quit, XSIG_BTNRELEASE);
}
void 
xpc_press_next(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_next, XSIG_BTNPRESS);
}
void 
xpc_release_next(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_next, XSIG_BTNRELEASE);
}
void 
xpc_press_prev(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_prev, XSIG_BTNPRESS);
}

void 
xpc_release_prev(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_prev, XSIG_BTNRELEASE);
}
void 
xpc_press_up(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_up, XSIG_BTNPRESS);
}
void 
xpc_release_up(struct xpopdic_control *cp)
{
	xpc_btn_send(cp, cp->xpc->xo_up, XSIG_BTNRELEASE);
}

static struct dictnode *
xpc_word_lookup(booklist books, const char *word, struct dictnode *current)
{
	char w[256];
	int  wlen = strlen(word);
	struct dictnode *next = dl_lookup(books, word, current);

	if(next || wlen < 4)
		return next;

	strncpy(w, word, 256);

	/*  
	 *  the plural of a naun, the past tense of a verb or
	 * the third person singular.
	 *	"~s", "~d", "~es", "~ed", "~ies", "~ied"
	 */
	if(word[wlen-1] == 's' || word[wlen-1] == 'd'){
		w[wlen-1] = '\0';
		next = dl_lookup(books, w, current);
		if(next)
			return next;
	}

	if(!strcmp(&word[wlen-2], "es") || 
	   !strcmp(&word[wlen-2], "ed")){
		w[wlen-2] = '\0';
		/* "tagged" or "tanned" */
		if(word[wlen-1] == 'd' && 
		   word[wlen-3] == word[wlen-4])
			w[wlen-3] = '\0';

		if((next = dl_lookup(books, w, current)) != NULL)
			return next;
	}

	if(!strcmp(&word[wlen-3], "ies") || 
	   !strcmp(&word[wlen-3], "ied")){
		w[wlen-3] = 'y';
		w[wlen-2] = '\0';
		if((next = dl_lookup(books, w, current)) != NULL)
			return next;
	}

	/* "owner" or "drier" */
	if(!strcmp(&word[wlen-2], "er")){
		w[wlen-1] = '\0';
		if((next = dl_lookup(books, w, current)) != NULL)
			return next;

		w[wlen-2] = '\0';

		/* "runner" */
		if(word[wlen-3] == word[wlen-4])
			w[wlen-3] = '\0';

		if((next = dl_lookup(books, w, current)) != NULL)
			return next;

		if(word[wlen-3] == 'i'){
			w[wlen-3] = '\0';
			if((next = dl_lookup(books, w, current)) != NULL)
				return next;
		}
	}

	/* "~ing" */
	if(!strcmp(&word[wlen-3], "ing")){
		w[wlen-3] = '\0';

		/* "running" */
		if(wlen > 6 && word[wlen-4] == word[wlen-5])
			w[wlen-4] = '\0';
		if((next = dl_lookup(books, w, current)) != NULL)
			return next;
	}

	return NULL;
}
